#include <iMesh_extensions.h>
#include <iRel.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define UNUSED __attribute__((unused))

/* To reuse some code that's usually parallel */
#define MPI_Comm_rank(a,b) (*(b)=0,0)

/* iRel has different error reporting, this hack is for uniform reporting */
extern struct iBase_Error iMesh_LAST_ERROR;
extern struct iBase_Error iRel_LAST_ERROR;

static void ReportErrorITAPS(int err,const char *file,int line,const char *func,const char *iface,const struct iBase_Error *estruct)
{
  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  fflush(stdout);
  fprintf(stderr,"[%d] %s:%d in %s() %s error %d(%d) description:\n%s\n",rank,file,line,func,iface,
          err,estruct->error_type,estruct->description);
}

static void ReportError(int err,const char *file,int line,const char *func)
{
  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  fflush(stdout);
  fprintf(stderr,"[%d] %s:%d in %s() [%d]\n",rank,file,line,func,err);
}

#define MCHK(err) if (err) { ReportErrorITAPS(err,__FILE__,__LINE__,__func__,"iMesh",&iMesh_LAST_ERROR); return err; }
#define GCHK(err) if (err) {                                            \
  int _err;                                                             \
  struct iBase_Error _last;                                              \
  iGeom_getDescription(geom,_last.description,&_err,sizeof(_last.description)); \
  _last.error_type = err;                                                \
  ReportErrorITAPS(err,__FILE__,__LINE__,__func__,"iGeom",&_last); return err; }
#define RCHK(err) if (err) { ReportErrorITAPS(err,__FILE__,__LINE__,__func__,"iRel", &iRel_LAST_ERROR); return err; }
#define CHK(err) if (err) { ReportError(err,__FILE__,__LINE__,__func__); return err; }

static int MeshCube(iMesh_Instance mesh,iBase_EntitySetHandle *facesets)
{
  static const double coords[3*8] = {-1,-1,-1, 1,-1,-1, 1,1,-1, -1,1,-1,
                                     -1,-1,1,  1,-1,1,  1,1,1,  -1,1,1};
  static const int ifconn[4*6] = {0,1,5,4, 1,2,6,5, 2,3,7,6, 3,0,4,7, 0,1,2,3, 4,5,6,7};
  iBase_EntityHandle verts[8],cube,*verts_p = verts,fconn[4*6],faces[6],*faces_p = faces;
  iBase_EntitySetHandle wallset,topset,bottomset,senseset;
  iBase_TagHandle bdyTag,senseTag;
  int err,verts_a=8,verts_s,faces_a = 6,faces_s,status[6],*status_p = status,status_a = 6,status_s;

  iMesh_createVtxArr(mesh,8,iBase_INTERLEAVED,coords,3*8,&verts_p,&verts_a,&verts_s,&err);MCHK(err);
  iMesh_createEnt(mesh,iMesh_HEXAHEDRON,verts,verts_s,&cube,&status[0],&err);MCHK(err);
  for (int i=0; i<4*6; i++) fconn[i] = verts[ifconn[i]];
  iMesh_createEntArr(mesh,iMesh_QUADRILATERAL,fconn,4*6,&faces_p,&faces_a,&faces_s,&status_p,&status_a,&status_s,&err);MCHK(err);
  for (int i=0; i<6; i++) {
    iMesh_createEntSet(mesh,0,&facesets[i],&err);MCHK(err);
    iMesh_addEntToSet(mesh,faces[i],facesets[i],&err);MCHK(err);
  }
  iMesh_getTagHandle(mesh,"NEUMANN_SET",&bdyTag,&err,sizeof("NEUMANN_SET"));MCHK(err);
  iMesh_createTag(mesh,"SENSE",1,iBase_INTEGER,&senseTag,&err,sizeof "SENSE");MCHK(err);
  iMesh_createEntSet(mesh,0,&wallset,&err);MCHK(err);
  iMesh_createEntSet(mesh,0,&topset,&err);MCHK(err);
  iMesh_createEntSet(mesh,0,&bottomset,&err);MCHK(err);
  iMesh_setEntSetIntData(mesh,wallset,bdyTag,100,&err);MCHK(err);
  iMesh_setEntSetIntData(mesh,topset,bdyTag,200,&err);MCHK(err);
  iMesh_setEntSetIntData(mesh,bottomset,bdyTag,300,&err);MCHK(err);
  for (int i=0; i<4; i++) {iMesh_addEntSet(mesh,facesets[i],wallset,&err);MCHK(err);}
  iMesh_addEntSet(mesh,facesets[5],topset,&err);MCHK(err);
  iMesh_addEntSet(mesh,facesets[4],bottomset,&err);MCHK(err);

  iMesh_createEntSet(mesh,0,&senseset,&err);MCHK(err);
  iMesh_addEntSet(mesh,facesets[4],senseset,&err);MCHK(err);
  iMesh_setEntSetIntData(mesh,senseset,senseTag,-1,&err);MCHK(err);
  iMesh_addEntSet(mesh,senseset,bottomset,&err);MCHK(err);
  return 0;
}

static int PrintBoundingBox(iGeom_Instance geom,iBase_EntityHandle gent,const char *name)
{
  int err;
  double x0,y0,z0,x1,y1,z1;
  iGeom_getEntBoundBox(geom,gent,&x0,&y0,&z0,&x1,&y1,&z1,&err);GCHK(err);
  printf("Geom `%s' bounding box (%f,%f)x(%f,%f)x(%f,%f)\n",name,x0,x1,y0,y1,z0,z1);
  return 0;
}

static int CreateGeomAndRelate(iRel_Instance assoc,iRel_RelationHandle rel,iGeom_Instance geom,iMesh_Instance UNUSED mesh,iBase_EntitySetHandle *facesets,int assoc_with_volume)
{
  iBase_EntityHandle cube,gfaces[6],*gfaces_p = gfaces;
  int err,gfaces_a = 6,gfaces_s;

  iGeom_createBrick(geom,2,2,2,&cube,&err);GCHK(err);
  err = PrintBoundingBox(geom,cube,"new cube");CHK(err);
  iGeom_getEntAdj(geom,cube,2,&gfaces_p,&gfaces_a,&gfaces_s,&err);GCHK(err);
  for (int i=0; i<6; i++) {
    char name[20];
    sprintf(name,"new face %d",i);
    err = PrintBoundingBox(geom,gfaces[i],name);CHK(err);
  }
  if (assoc_with_volume) {
    /* Not what we want */
    for (int i=0; i<6; i++) {
      iRel_setEntSetAssociation(assoc,rel,cube,facesets[i],&err);RCHK(err);
    }
  } else {
    iRel_setEntSetAssociation(assoc,rel,gfaces[0],facesets[3],&err);RCHK(err);
    iRel_setEntSetAssociation(assoc,rel,gfaces[1],facesets[1],&err);RCHK(err);
    iRel_setEntSetAssociation(assoc,rel,gfaces[2],facesets[0],&err);RCHK(err);
    iRel_setEntSetAssociation(assoc,rel,gfaces[3],facesets[2],&err);RCHK(err);
    iRel_setEntSetAssociation(assoc,rel,gfaces[4],facesets[4],&err);RCHK(err);
    iRel_setEntSetAssociation(assoc,rel,gfaces[5],facesets[5],&err);RCHK(err);
  }
  return 0;
}

static int MeshSetGetEntityCounts(iMesh_Instance instance,iBase_EntitySetHandle set,int count[],int rcount[])
{
  int err;
  for (int type=0; type<4; type++) {
    iMesh_getNumOfType(instance,set,type,&count[type],&err);MCHK(err);
    iMesh_getNumOfTypeRec(instance,set,type,1,&rcount[type],&err);MCHK(err);
  }
  return 0;
}
static int GeomSetGetEntityCounts(iGeom_Instance instance,iBase_EntitySetHandle set,int count[])
{
  int err;
  for (int type=0; type<4; type++) {
    iGeom_getNumOfType(instance,set,type,&count[type],&err);MCHK(err);
  }
  return 0;
}

static int EnumerateSubset(iMesh_Instance mesh,int hier[],int level,iBase_EntitySetHandle set)
{
  static const char *type_name[4] = {"INTEGER","DOUBLE","ENT_HANDLE","BYTES"};
  char name[256],value_buf[1024],*values = value_buf;
  int values_a = sizeof(value_buf),values_s;
  int err,rank,tags_a=0,tags_s,subsets_a=0,subsets_s,n[4],nr[4],tagtype,tagsize;
  iBase_EntitySetHandle *subsets = 0;
  iBase_TagHandle *tags = 0;
  err = MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  err = MeshSetGetEntityCounts(mesh,set,n,nr);CHK(err);
  iMesh_getEntSets(mesh,set,1,&subsets,&subsets_a,&subsets_s,&err);MCHK(err);
  printf("[%d] Subset %p {%3d,%2d,%2d}(%d)  EntTypes (%3d,%3d,%3d,%3d) Contains %d subsets\n",
         rank,(void*)set,hier[0],hier[1],hier[2],level,n[0],n[1],n[2],n[3],subsets_s);
  iMesh_getAllEntSetTags(mesh,set,&tags,&tags_a,&tags_s,&err);MCHK(err);
  for (int i=0; i<tags_s; i++) {
    iMesh_getTagName(mesh,tags[i],name,&err,sizeof(name));MCHK(err);
    iMesh_getTagType(mesh,tags[i],&tagtype,&err);MCHK(err);
    iMesh_getTagSizeValues(mesh,tags[i],&tagsize,&err);MCHK(err);
    printf("[%d] Subset %p {%3d,%2d,%2d}(%d)  Tag %d \"%s\" %s[%d] :",
           rank,(void*)set,hier[0],hier[1],hier[2],level,i,name,type_name[tagtype],tagsize);
    iMesh_getEntSetData(mesh,set,tags[i],&values,&values_a,&values_s,&err);MCHK(err);
    for (int j=0; j<tagsize; j++) {
      switch (tagtype) {
        case iBase_INTEGER: printf(" %d",((int*)values)[j]); break;
        case iBase_DOUBLE: printf(" %f",((double*)values)[j]); break;
        case iBase_ENTITY_HANDLE: printf(" %p",((void**)values)[j]); break;
        case iBase_BYTES: {
          int v = ((char*)values)[j];
          printf(isprint(v)?"%c":":%hhx",v);
        } break;
        default: return 1;
      }
    }
    printf("\n");
  }
  for (int i=0; i<subsets_s; i++) {
    if (level > 2) {
      printf("[%d] Cannot explore subset %p(%d)\n",rank,(void*)subsets[i],i);
      continue;
    }
    hier[level] = i+1;
    err = EnumerateSubset(mesh,hier,level+1,subsets[i]);CHK(err);
  }
  hier[level] = 0;
  return 0;
}

static int EnumerateNeumannSet(iMesh_Instance mesh,iBase_TagHandle tag,iBase_EntitySetHandle set)
{
  iBase_EntitySetHandle *subsets=0;
  int err,n[4],nr[4],subsets_a=0,subsets_s,value,rank,hier[10]={0};
  err = MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  iMesh_getEntSetIntData(mesh,set,tag,&value,&err);MCHK(err);
  iMesh_getEntSets(mesh,set,1,&subsets,&subsets_a,&subsets_s,&err);MCHK(err);
  err = MeshSetGetEntityCounts(mesh,set,n,nr);CHK(err);
  printf("[%d] NEUMANN_SET %p(%3d)  EntTypes (%3d/%3d,%3d/%3d,%3d/%3d,%3d/%3d) contains %d subsets\n",rank,(void*)set,value,n[0],nr[0],n[1],nr[1],n[2],nr[2],n[3],nr[3],subsets_s);
  hier[0] = value;
  for (int i=0; i<subsets_s; i++) {
    hier[1] = i+1;              /* Count from 1 */
    err = EnumerateSubset(mesh,hier,1,subsets[i]);CHK(err);
  }
  return 0;
}

static int ExamineRelations(iRel_Instance assoc,iRel_RelationHandle rel,iGeom_Instance geom,iMesh_Instance mesh,int nsetno)
{
  const char nsetname[] = "NEUMANN_SET",ssetname[] = "SENSE";
  int err,rank,*tagval = &nsetno,nsets_a=0,nsets_s;
  iBase_TagHandle tag,sensetag;
  iBase_EntitySetHandle *nsets=0,root,emptyset;

  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  iMesh_getRootSet(mesh,&root,&err);MCHK(err);
  iMesh_getTagHandle(mesh,nsetname,&tag,&err,sizeof(nsetname));MCHK(err);
  iMesh_getTagHandle(mesh,ssetname,&sensetag,&err,sizeof(ssetname));MCHK(err);
  if (err == iBase_TAG_NOT_FOUND) { /* Create the tag so we don't need to deal with it not existing */
    iMesh_createTag(mesh,ssetname,1,iBase_INTEGER,&sensetag,&err,sizeof(ssetname));
  } else MCHK(err);
  iMesh_createEntSet(mesh,0,&emptyset,&err);MCHK(err); /* Create an empty set, also to avoid excessive errors */
  if (1) {/* Find all Neumann sets */
    iMesh_getEntSetsByTagsRec(mesh,root,&tag,0,1,0,&nsets,&nsets_a,&nsets_s,&err);MCHK(err);
    printf("[%d] Found %d Neumann sets\n",rank,nsets_s);
    for (int i=0; i<nsets_s; i++) {
      err = EnumerateNeumannSet(mesh,tag,nsets[i]);CHK(err);
    }
  }
  /* Now just get the requested Neumann sets */
  iMesh_getEntSetsByTagsRec(mesh,root,&tag,(const char**)&tagval,1,0,&nsets,&nsets_a,&nsets_s,&err);MCHK(err);
  if (nsets_s > 1) { printf("[%d] Weird, more than one set with given value of NEUMANN_SET.\n",rank); }
  for (int i=0; i<nsets_s; i++) {
    iBase_EntitySetHandle sense_set,*sense_set_p = &sense_set,subsets[100],*subsets_p = subsets;
    int sense_set_a = 1,sense_set_s,subsets_a = 100,subsets_s,value = -1,*value_p=&value;
    iMesh_getEntSetsByTagsRec(mesh,nsets[i],&sensetag,(const char**)&value_p,1,0,&sense_set_p,&sense_set_a,&sense_set_s,&err);MCHK(err);
    if (!sense_set_s) { sense_set = emptyset; }
    iMesh_getEntSets(mesh,nsets[i],1,&subsets_p,&subsets_a,&subsets_s,&err);MCHK(err);
    for (int j=0; j<subsets_s; j++) {
      iBase_EntityHandle gent=0;
      int gtype = -1;
      iBase_EntityHandle faces[1000],*faces_p = faces,verts[1000],*verts_p = verts;
      double coords[3000],*coords_p = coords;
      int faces_a = 1000,faces_s,verts_a=1000,verts_s,coords_a=3000,coords_s;
      {
        double x0,x1,y0,y1,z0,z1;
        gent = 0; gtype = -1;
        x0=x1=y0=y1=z0=z1=0;
        iRel_getSetEntAssociation(assoc,rel,subsets[j],1,&gent,&err); if (err) { gent = 0; } //RCHK(err);
        if (gent) {
          iGeom_getEntType(geom,gent,&gtype,&err); if (err) gtype = -2; //GCHK(err);
          iGeom_getEntBoundBox(geom,gent,&x0,&y0,&z0,&x1,&y1,&z1,&err);GCHK(err);
        }
        printf("[%d] NEUMANN_SET %d.%d.%d is associated with geometry %p of type %d bounded by (%f,%f)x(%f,%f)x(%f,%f)\n",rank,nsetno,i,j,(void*)gent,gtype,x0,x1,y0,y1,z0,z1);
      }
      iMesh_getEntities(mesh,subsets[j],iBase_FACE,iMesh_ALL_TOPOLOGIES,&faces_p,&faces_a,&faces_s,&err);MCHK(err);
      iMesh_getEntities(mesh,subsets[j],iBase_VERTEX,iMesh_ALL_TOPOLOGIES,&verts_p,&verts_a,&verts_s,&err);MCHK(err);
      iMesh_getVtxArrCoords(mesh,verts,verts_s,iBase_INTERLEAVED,&coords_p,&coords_a,&coords_s,&err);MCHK(err);
      for (int k=0; k<verts_s; k++) {
        double nrml[3],*nrml_p = nrml,xyz[3],*xyz_p = xyz;
        int nrml_a=3,nrml_s,xyz_a=3,xyz_s;
        iGeom_getArrNrmlPlXYZ(geom,&gent,1,iBase_INTERLEAVED,&coords[3*k],3,&xyz_p,&xyz_a,&xyz_s,&nrml_p,&nrml_a,&nrml_s,&err);GCHK(err);
        printf("[%d] NEUMANN_SET %d.%d.%d vertex %d (%f,%f,%f) is associated with geometry %p of type %d\n"
               "\t nearest (%f,%f,%f) normal (%f,%f,%f)\n",rank,nsetno,i,j,k,coords[0],coords[1],coords[2],(void*)gent,gtype,
               xyz[0],xyz[1],xyz[2],nrml[0],nrml[1],nrml[2]);
      }
    }
  }
  free(nsets);
  return 0;
}


static int ExamineGeometry(iGeom_Instance geom)
{
  int err,rank,n[4],ns;
  iBase_EntitySetHandle root;
  double xs[3],xm[3];

  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  iGeom_getRootSet(geom,&root,&err);GCHK(err);
  iGeom_getBoundBox(geom,&xs[0],&xs[1],&xs[2],&xm[0],&xm[1],&xm[2],&err);GCHK(err);
  printf("[%d] Geom: Bounding box (%f,%f) (%f,%f) (%f,%f)\n",rank,xs[0],xm[0],xs[1],xm[1],xs[2],xm[2]);
  err = GeomSetGetEntityCounts(geom,root,n);CHK(err);
  iGeom_getNumEntSets(geom,root,1,&ns,&err);GCHK(err);
  printf("[%d] Geom: Total num by type (%d,%d,%d,%d)  num sets %d\n",rank,n[0],n[1],n[2],n[3],ns);
  if (1) {                      /* Investigate faces */
    iBase_EntityHandle faces[1000],*faces_p = faces;
    int faces_a=1000,faces_s;
    char facetype[100] = {0};
    iGeom_getEntities(geom,root,iBase_FACE,&faces_p,&faces_a,&faces_s,&err);GCHK(err);
    for (int i=0; i<faces_s; i++) {
      int facetype_s = sizeof(facetype) - 1;/* Otherwise CGM can write off the end */
      iGeom_getFaceType(geom,faces[i],facetype,&err,&facetype_s);GCHK(err);
      printf("[%d] Face %d %p has type %s\n",rank,i,(void*)faces[i],facetype);
      iGeom_getEntBoundBox(geom,faces[i],&xs[0],&xs[1],&xs[2],&xm[0],&xm[1],&xm[2],&err);GCHK(err);
      printf("[%d] Face %d %p Bounding box (%f,%f) (%f,%f) (%f,%f)\n",rank,i,(void*)faces[i],xs[0],xm[0],xs[1],xm[1],xs[2],xm[2]);
    }
  }
  return 0;
}

int main(int argc,char *argv[])
{
  static const char
    mesh_filename[] = "cube.h5m",
    mesh_options_new[] = "",
    mesh_options_load[] = "",
    mesh_options_save[] = "",
    geom_filename[] = "cube.brep",
    geom_options_new[] = ";ENGINE=OCC;",
    geom_options_load[] = ";TYPE=OCC;",
    geom_options_save[] = ";TYPE=OCC;",
    rel_options_new[] = "";
  
  iMesh_Instance mesh;
  iGeom_Instance geom;
  iRel_Instance assoc;
  iRel_RelationHandle rel;
  iBase_EntitySetHandle facesets[6];
  int err;

  if (argc < 2 || argc > 3) goto usage;
  iMesh_newMesh(mesh_options_new,&mesh,&err,sizeof mesh_options_new);MCHK(err);
  iGeom_newGeom(geom_options_new,&geom,&err,sizeof geom_options_new);GCHK(err);
  iRel_newAssoc(rel_options_new,&assoc,&err,sizeof rel_options_new);RCHK(err);
  iRel_createAssociation(assoc,geom,0,iRel_IGEOM_IFACE,mesh,1,iRel_IMESH_IFACE,&rel,&err);RCHK(err);

  if (!strcmp(argv[1],"-create")) {
    int assoc_with_volume = 0;
    if (argc == 3) {
      if (!strcmp(argv[2],"-assoc_with_volume")) assoc_with_volume = 1;
      else goto usage;
    }
    err = MeshCube(mesh,facesets);CHK(err);
    err = CreateGeomAndRelate(assoc,rel,geom,mesh,facesets,assoc_with_volume);CHK(err);
    iMesh_save(mesh,0,mesh_filename,mesh_options_save,&err,sizeof mesh_filename,sizeof mesh_options_save);MCHK(err);
    iGeom_save(geom,geom_filename,geom_options_save,&err,sizeof geom_filename,sizeof geom_options_save);GCHK(err);
  } else if (!strcmp(argv[1],"-read")) {
    iMesh_load(mesh,0,mesh_filename,mesh_options_load,&err,sizeof mesh_filename,sizeof mesh_options_load);MCHK(err);
    iGeom_load(geom,geom_filename,geom_options_load,&err,sizeof geom_filename,sizeof geom_options_load);GCHK(err);
    iRel_inferAllAssociations(assoc,rel,&err);RCHK(err);
  } else goto usage;
  err = ExamineGeometry(geom);CHK(err);
  err = ExamineRelations(assoc,rel,geom,mesh,100);CHK(err);
  return 0;

usage:
  fprintf(stderr,
          "Usage: %s -create [-assoc_with_volume]     (writes mesh and geometry)\n"
          "       %s -read                            (read mesh and geometry, associate)\n",
          argv[0],argv[0]);
  return 1;
}
